@@ -4,6 +4,7 @@
 #include <xf86drm.h>
 #include <xf86drmMode.h>
 #include "backend/drm/drm.h"
+#include "backend/drm/fb.h"
 #include "backend/drm/iface.h"
 #include "backend/drm/util.h"
 
@@ -33,11 +34,12 @@ static bool legacy_fb_props_match(struct wlr_drm_fb *fb1,
 	return true;
 }
 
-static bool legacy_crtc_test(struct wlr_drm_connector *conn,
-		const struct wlr_drm_connector_state *state) {
+static bool legacy_crtc_test(const struct wlr_drm_connector_state *state,
+		bool modeset) {
+	struct wlr_drm_connector *conn = state->connector;
 	struct wlr_drm_crtc *crtc = conn->crtc;
 
-	if ((state->base->committed & WLR_OUTPUT_STATE_BUFFER) && !state->modeset) {
+	if ((state->base->committed & WLR_OUTPUT_STATE_BUFFER) && !modeset) {
 		struct wlr_drm_fb *pending_fb = state->primary_fb;
 
 		struct wlr_drm_fb *prev_fb = crtc->primary->queued_fb;
@@ -57,16 +59,9 @@ static bool legacy_crtc_test(struct wlr_drm_connector *conn,
 	return true;
 }
 
-static bool legacy_crtc_commit(struct wlr_drm_connector *conn,
-		const struct wlr_drm_connector_state *state,
-		struct wlr_drm_page_flip *page_flip, uint32_t flags, bool test_only) {
-	if (!legacy_crtc_test(conn, state)) {
-		return false;
-	}
-	if (test_only) {
-		return true;
-	}
-
+static bool legacy_crtc_commit(const struct wlr_drm_connector_state *state,
+		struct wlr_drm_page_flip *page_flip, uint32_t flags, bool modeset) {
+	struct wlr_drm_connector *conn = state->connector;
 	struct wlr_drm_backend *drm = conn->backend;
 	struct wlr_output *output = &conn->output;
 	struct wlr_drm_crtc *crtc = conn->crtc;
@@ -82,7 +77,7 @@ static bool legacy_crtc_commit(struct wlr_drm_connector *conn,
 		fb_id = state->primary_fb->id;
 	}
 
-	if (state->modeset) {
+	if (modeset) {
 		uint32_t *conns = NULL;
 		size_t conns_len = 0;
 		drmModeModeInfo *mode = NULL;
@@ -115,10 +110,11 @@ static bool legacy_crtc_commit(struct wlr_drm_connector *conn,
 	}
 
 	if (state->base->committed & WLR_OUTPUT_STATE_ADAPTIVE_SYNC_ENABLED) {
-		if (!drm_connector_supports_vrr(conn)) {
+		if (state->base->adaptive_sync_enabled && !output->adaptive_sync_supported) {
 			return false;
 		}
-		if (drmModeObjectSetProperty(drm->fd, crtc->id, DRM_MODE_OBJECT_CRTC,
+		if (crtc->props.vrr_enabled != 0 &&
+				drmModeObjectSetProperty(drm->fd, crtc->id, DRM_MODE_OBJECT_CRTC,
 				crtc->props.vrr_enabled,
 				state->base->adaptive_sync_enabled) != 0) {
 			wlr_drm_conn_log_errno(conn, WLR_ERROR,
@@ -133,7 +129,7 @@ static bool legacy_crtc_commit(struct wlr_drm_connector *conn,
 	}
 
 	if (cursor != NULL && drm_connector_is_cursor_visible(conn)) {
-		struct wlr_drm_fb *cursor_fb = get_next_cursor_fb(conn);
+		struct wlr_drm_fb *cursor_fb = state->cursor_fb;
 		if (cursor_fb == NULL) {
 			wlr_drm_conn_log(conn, WLR_DEBUG, "Failed to acquire cursor FB");
 			return false;
@@ -184,6 +180,32 @@ static bool legacy_crtc_commit(struct wlr_drm_connector *conn,
 	return true;
 }
 
+static bool legacy_commit(struct wlr_drm_backend *drm,
+		const struct wlr_drm_device_state *state,
+		struct wlr_drm_page_flip *page_flip, uint32_t flags,
+		bool test_only) {
+	for (size_t i = 0; i < state->connectors_len; i++) {
+		const struct wlr_drm_connector_state *conn_state = &state->connectors[i];
+		if (!legacy_crtc_test(conn_state, state->modeset)) {
+			return false;
+		}
+	}
+
+	if (test_only) {
+		return true;
+	}
+
+	for (size_t i = 0; i < state->connectors_len; i++) {
+		const struct wlr_drm_connector_state *conn_state = &state->connectors[i];
+		if (!legacy_crtc_commit(conn_state, page_flip, flags,
+				state->modeset)) {
+			return false;
+		}
+	}
+
+	return true;
+}
+
 static void fill_empty_gamma_table(size_t size,
 		uint16_t *r, uint16_t *g, uint16_t *b) {
 	assert(0xFFFF < UINT64_MAX / (size - 1));
@@ -226,6 +248,20 @@ bool drm_legacy_crtc_set_gamma(struct wlr_drm_backend *drm,
 	return true;
 }
 
+static bool legacy_reset(struct wlr_drm_backend *drm) {
+	bool ok = true;
+	for (size_t i = 0; i < drm->num_crtcs; i++) {
+		struct wlr_drm_crtc *crtc = &drm->crtcs[i];
+		if (drmModeSetCrtc(drm->fd, crtc->id, 0, 0, 0, NULL, 0, NULL) != 0) {
+			wlr_log_errno(WLR_ERROR, "Failed to disable CRTC %"PRIu32,
+				crtc->id);
+			ok = false;
+		}
+	}
+	return ok;
+}
+
 const struct wlr_drm_interface legacy_iface = {
-	.crtc_commit = legacy_crtc_commit,
+	.commit = legacy_commit,
+	.reset = legacy_reset,
 };
