@@ -1,4 +1,3 @@
-#define _POSIX_C_SOURCE 200809L
 #include <assert.h>
 #include <fcntl.h>
 #include <limits.h>
@@ -132,7 +131,7 @@ static int x11_event(int fd, uint32_t mask, void *data) {
 		if (mask & WL_EVENT_ERROR) {
 			wlr_log(WLR_ERROR, "Failed to read from X11 server");
 		}
-		wl_display_terminate(x11->wl_display);
+		wlr_backend_destroy(&x11->backend);
 		return 0;
 	}
 
@@ -145,7 +144,7 @@ static int x11_event(int fd, uint32_t mask, void *data) {
 	int ret = xcb_connection_has_error(x11->xcb);
 	if (ret != 0) {
 		wlr_log(WLR_ERROR, "X11 connection error (%d)", ret);
-		wl_display_terminate(x11->wl_display);
+		wlr_backend_destroy(&x11->backend);
 	}
 
 	return 0;
@@ -192,7 +191,7 @@ static void backend_destroy(struct wlr_backend *backend) {
 	if (x11->event_source) {
 		wl_event_source_remove(x11->event_source);
 	}
-	wl_list_remove(&x11->display_destroy.link);
+	wl_list_remove(&x11->event_loop_destroy.link);
 
 	wlr_drm_format_set_finish(&x11->primary_dri3_formats);
 	wlr_drm_format_set_finish(&x11->primary_shm_formats);
@@ -230,9 +229,8 @@ bool wlr_backend_is_x11(struct wlr_backend *backend) {
 	return backend->impl == &backend_impl;
 }
 
-static void handle_display_destroy(struct wl_listener *listener, void *data) {
-	struct wlr_x11_backend *x11 =
-		wl_container_of(listener, x11, display_destroy);
+static void handle_event_loop_destroy(struct wl_listener *listener, void *data) {
+	struct wlr_x11_backend *x11 = wl_container_of(listener, x11, event_loop_destroy);
 	backend_destroy(&x11->backend);
 }
 
@@ -263,11 +261,13 @@ static int query_dri3_drm_fd(struct wlr_x11_backend *x11) {
 	xcb_dri3_open_reply_t *open_reply =
 		xcb_dri3_open_reply(x11->xcb, open_cookie, NULL);
 	if (open_reply == NULL) {
+		wlr_log(WLR_ERROR, "Failed to open DRI3");
 		return -1;
 	}
 
 	int *open_fds = xcb_dri3_open_reply_fds(x11->xcb, open_reply);
 	if (open_fds == NULL) {
+		wlr_log(WLR_ERROR, "xcb_dri3_open_reply_fds() failed");
 		free(open_reply);
 		return -1;
 	}
@@ -279,10 +279,12 @@ static int query_dri3_drm_fd(struct wlr_x11_backend *x11) {
 
 	int flags = fcntl(drm_fd, F_GETFD);
 	if (flags < 0) {
+		wlr_log_errno(WLR_ERROR, "Failed to get DRM FD flags");
 		close(drm_fd);
 		return -1;
 	}
 	if (fcntl(drm_fd, F_SETFD, flags | FD_CLOEXEC) < 0) {
+		wlr_log_errno(WLR_ERROR, "Failed to set DRM FD flags");
 		close(drm_fd);
 		return -1;
 	}
@@ -290,6 +292,7 @@ static int query_dri3_drm_fd(struct wlr_x11_backend *x11) {
 	if (drmGetNodeTypeFromFd(drm_fd) != DRM_NODE_RENDER) {
 		char *render_name = drmGetRenderDeviceNameFromFd(drm_fd);
 		if (render_name == NULL) {
+			wlr_log(WLR_ERROR, "Failed to get DRM render node name from DRM FD");
 			close(drm_fd);
 			return -1;
 		}
@@ -297,6 +300,7 @@ static int query_dri3_drm_fd(struct wlr_x11_backend *x11) {
 		close(drm_fd);
 		drm_fd = open(render_name, O_RDWR | O_CLOEXEC);
 		if (drm_fd < 0) {
+			wlr_log_errno(WLR_ERROR, "Failed to open DRM render node '%s'", render_name);
 			free(render_name);
 			return -1;
 		}
@@ -391,7 +395,7 @@ static void x11_get_argb32(struct wlr_x11_backend *x11) {
 	free(reply);
 }
 
-struct wlr_backend *wlr_x11_backend_create(struct wl_display *display,
+struct wlr_backend *wlr_x11_backend_create(struct wl_event_loop *loop,
 		const char *x11_display) {
 	wlr_log(WLR_INFO, "Creating X11 backend");
 
@@ -401,7 +405,7 @@ struct wlr_backend *wlr_x11_backend_create(struct wl_display *display,
 	}
 
 	wlr_backend_init(&x11->backend, &backend_impl);
-	x11->wl_display = display;
+	x11->event_loop = loop;
 	wl_list_init(&x11->outputs);
 
 	x11->xcb = xcb_connect(x11_display, NULL);
@@ -561,9 +565,8 @@ struct wlr_backend *wlr_x11_backend_create(struct wl_display *display,
 	free(xi_reply);
 
 	int fd = xcb_get_file_descriptor(x11->xcb);
-	struct wl_event_loop *ev = wl_display_get_event_loop(display);
 	uint32_t events = WL_EVENT_READABLE | WL_EVENT_ERROR | WL_EVENT_HANGUP;
-	x11->event_source = wl_event_loop_add_fd(ev, fd, events, x11_event, x11);
+	x11->event_source = wl_event_loop_add_fd(loop, fd, events, x11_event, x11);
 	if (!x11->event_source) {
 		wlr_log(WLR_ERROR, "Could not create event source");
 		goto error_display;
@@ -600,7 +603,7 @@ struct wlr_backend *wlr_x11_backend_create(struct wl_display *display,
 
 	if (!query_formats(x11)) {
 		wlr_log(WLR_ERROR, "Failed to query supported DRM formats");
-		return false;
+		goto error_event;
 	}
 
 	x11->drm_fd = -1;
@@ -610,7 +613,8 @@ struct wlr_backend *wlr_x11_backend_create(struct wl_display *display,
 		x11->drm_fd = query_dri3_drm_fd(x11);
 		if (x11->drm_fd < 0) {
 			wlr_log(WLR_ERROR, "Failed to query DRI3 DRM FD");
-			goto error_event;
+			wlr_log(WLR_INFO, "Disabling DMA-BUF support");
+			x11->have_dri3 = false;
 		}
 	}
 
@@ -637,15 +641,15 @@ struct wlr_backend *wlr_x11_backend_create(struct wl_display *display,
 #if HAVE_XCB_ERRORS
 	if (xcb_errors_context_new(x11->xcb, &x11->errors_context) != 0) {
 		wlr_log(WLR_ERROR, "Failed to create error context");
-		return false;
+		goto error_event;
 	}
 #endif
 
 	wlr_keyboard_init(&x11->keyboard, &x11_keyboard_impl,
 		x11_keyboard_impl.name);
 
-	x11->display_destroy.notify = handle_display_destroy;
-	wl_display_add_destroy_listener(display, &x11->display_destroy);
+	x11->event_loop_destroy.notify = handle_event_loop_destroy;
+	wl_event_loop_add_destroy_listener(loop, &x11->event_loop_destroy);
 
 	// Create an empty pixmap to be used as the cursor. The
 	// default GC foreground is 0, and that is what it will be
