@@ -72,6 +72,7 @@ enum atom_name {
 	NET_WM_WINDOW_TYPE_MENU,
 	NET_WM_WINDOW_TYPE_NOTIFICATION,
 	NET_WM_WINDOW_TYPE_SPLASH,
+	NET_WM_WINDOW_TYPE_DESKTOP,
 	DND_SELECTION,
 	DND_AWARE,
 	DND_STATUS,
@@ -134,6 +135,7 @@ struct wlr_xwm {
 	struct wl_listener compositor_new_surface;
 	struct wl_listener compositor_destroy;
 	struct wl_listener shell_v1_new_surface;
+	struct wl_listener shell_v1_destroy;
 	struct wl_listener seat_set_selection;
 	struct wl_listener seat_set_primary_selection;
 	struct wl_listener seat_start_drag;
@@ -161,4 +163,8 @@ char *xwm_get_atom_name(struct wlr_xwm *xwm, xcb_atom_t atom);
 bool xwm_atoms_contains(struct wlr_xwm *xwm, xcb_atom_t *atoms,
 	size_t num_atoms, enum atom_name needle);
 
+xcb_void_cookie_t xwm_send_event_with_size(xcb_connection_t *c,
+	uint8_t propagate, xcb_window_t destination,
+	uint32_t event_mask, const void *event, uint32_t length);
+
 #endif
